"""
pythonbible includes features for parsing texts for scripture references,
converting references into integer verse ids for efficient use/storage,
converting verse ids back into normalized references, and formatting
references as human-readable strings.
"""

__version__ = "0.1.2"

from .bible.osis.parser import OSISParser
from .books import Book
from .converter import (
    convert_reference_to_verse_ids,
    convert_references_to_verse_ids,
    convert_verse_ids_to_references,
)
from .errors import InvalidChapterError, InvalidVerseError
from .formatter import (
    format_scripture_references,
    format_scripture_text,
    format_single_reference,
    get_parser,
)
from .parser import (
    get_references,
    normalize_reference,
)
from .validator import (
    is_valid_book,
    is_valid_chapter,
    is_valid_reference,
    is_valid_verse,
    is_valid_verse_id,
)
from .verses import (
    get_book,
    get_book_chapter_verse,
    get_chapter,
    get_max_number_of_verses,
    get_verse,
    get_verse_id,
)
from .versions import Version



"""Contains the BibleParser generic parser class."""

from collections import OrderedDict


class BibleParser:
    """
    Parse files containing scripture text.

    BibleParser is a generic parser to provide common functionality for specific
    parsers (e.g. OSIS, USFM, USFX, etc.) for parsing scripture text.
    """

    def __init__(self, version):
        """
        Initialize the Bible parser with the version.

        :param version:
        """
        self.version = version


def sort_paragraphs(paragraphs):
    """
    Sort paragraphs of scripture text.

    Given a structured collection of paragraphs organized by book, chapter, and
    list of paragraphs, return that collection in an ordered dictionary with the
    books and chapters sorted appropriately. (Assume the list of paragraphs is
    already sorted appropriately.)

    :param paragraphs:
    :return: an OrderedDict(Book, OrderedDict(int, list(string)))
    """
    ordered_paragraphs = OrderedDict()

    book_keys = list(paragraphs.keys())
    book_keys.sort()

    for book in book_keys:
        chapters = paragraphs.get(book)
        ordered_chapters = OrderedDict()

        chapter_keys = list(chapters.keys())
        chapter_keys.sort()

        for chapter in chapter_keys:
            ordered_chapters[chapter] = chapters.get(chapter)

        ordered_paragraphs[book] = ordered_chapters

    return ordered_paragraphs



import pythonbible as bible


def test_format_scripture_references(
    normalized_references_complex, formatted_reference
):
    # Given a list of normalized references
    # When we format them into a reference string
    reference = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_references_null():
    # Given a null references object
    # When we attempt to format it into a reference string
    formatted_reference = bible.format_scripture_references(None)

    # Then the result is null
    assert formatted_reference is None


def test_format_scripture_references_sorting(
    normalized_references_complex, formatted_reference
):
    # Given a list of normalized references that are not in proper order
    normalized_references_complex.reverse()

    # When we format them into a reference string
    reference = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_text(verse_ids, html_scripture_text):
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text = bible.format_scripture_text(verse_ids)

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text


def test_format_scripture_text_non_html(verse_ids, non_html_scripture_text):
    # Given a list of verse ids
    # When we get the non html scripture text for those verse ids
    scripture_text = bible.format_scripture_text(verse_ids, format_type="text")

    # Then the scripture text is formatted correctly.
    assert scripture_text == non_html_scripture_text


def test_format_scripture_text_one_verse_per_paragraph(
    verse_ids, html_scripture_text_one_verse_per_paragraph
):
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text = bible.format_scripture_text(
        verse_ids, one_verse_per_paragraph=True
    )

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text_one_verse_per_paragraph



import pytest

import pythonbible as bible
from pythonbible.formatter import DEFAULT_PARSER


def test_get_scripture_passage_text(verse_ids_complex, kjv_passage):
    # Given a list of verse ids
    # When we get the scripture passage for those verses
    passage = DEFAULT_PARSER.get_scripture_passage_text(verse_ids_complex)

    # Then the scripture passage is correct.
    assert passage == kjv_passage


def test_get_scripture_passage_text_no_numbers(
    verse_ids_complex, kjv_passage_no_verse_numbers
):
    # Given a list of verse ids
    # When we get the scripture passage for those verses
    passage = DEFAULT_PARSER.get_scripture_passage_text(
        verse_ids_complex, include_verse_number=False
    )

    # Then the scripture passage is correct.
    assert passage == kjv_passage_no_verse_numbers


def test_get_asv_scripture_passage_text(verse_ids_complex, asv_passage):
    # Given a list of verse ids
    # When we get the ASV scripture passage for those verses
    parser = bible.OSISParser(bible.Version.AMERICAN_STANDARD)
    passage = parser.get_scripture_passage_text(verse_ids_complex)

    # Then the scripture passage is correct.
    assert passage == asv_passage


def test_get_scripture_passage_ellipsis():
    passage = DEFAULT_PARSER.get_scripture_passage_text(
        [
            1001003,
            1001005,
        ]
    )
    assert "..." in passage.get(bible.Book.GENESIS).get(1)[0]


def test_get_book_title():
    book_title = DEFAULT_PARSER.get_book_title(bible.Book.GENESIS)
    assert book_title == "The First Book of Moses, called Genesis"


def test_get_short_book_title():
    book_title = DEFAULT_PARSER.get_short_book_title(bible.Book.GENESIS)
    assert book_title == "Genesis"


def test_get_scripture_passage_null():
    assert DEFAULT_PARSER.get_scripture_passage_text(None) is None


def test_get_verse_text(verse_id, verse_text):
    # Given a verse id
    # When we get the scripture text for that verse id
    actual_text = DEFAULT_PARSER.get_verse_text(verse_id)

    # Then it is what we expect it to be.
    assert actual_text == verse_text


def test_get_verse_text_null():
    # Given a null verse id
    # When we attempt to get the scripture text for that verse id
    # Then it raise an InvalidVerseError
    with pytest.raises(bible.InvalidVerseError):
        DEFAULT_PARSER.get_verse_text(None)


def test_get_verse_text_invalid_verse(invalid_verse_id):
    # Given an invalid verse id
    # When we attempt to get the scripture text for that verse id
    # Then it raise an InvalidVerseError
    with pytest.raises(bible.InvalidVerseError):
        DEFAULT_PARSER.get_verse_text(invalid_verse_id)


def test_get_scripture_passage_one_verse_per_paragraph():
    """Test for https://github.com/avendesora/python-bible/issues/5."""
    # Given "Genesis 1" as a reference / verse ids
    references = bible.get_references("Genesis 1")
    verse_ids = bible.convert_references_to_verse_ids(references)

    # When we get the scripture passage with one verse per paragraph
    passage = DEFAULT_PARSER.get_scripture_passage_text(
        verse_ids, one_verse_per_paragraph=True
    )

    # Then it does not raise an exception
    assert passage is not None



from pythonbible.bible.osis.parser import OSISParser
from pythonbible.converter import (
    convert_references_to_verse_ids,
    convert_verse_ids_to_references,
)
from pythonbible.versions import Version

VERSION_MAP = {Version.AMERICAN_STANDARD: OSISParser, Version.KING_JAMES: OSISParser}

DEFAULT_VERSION = Version.KING_JAMES


def get_parser(**kwargs):
    version = kwargs.get("version", DEFAULT_VERSION)
    version_map = kwargs.get("version_map", VERSION_MAP)
    return version_map.get(version)(version)


DEFAULT_PARSER = get_parser()


# TODO - handle Psalms vs Psalm appropriately
# TODO - handle single chapter books appropriately (e.g. Obadiah 1-4 rather than Obadiah 1:1-4)
def format_scripture_references(references, **kwargs):
    """

    :param references: a list of normalized scripture references
    :return: a string version of the references formatted to be human-readable
    """
    if references is None:
        return None

    verse_ids = convert_references_to_verse_ids(references)
    verse_ids.sort()
    sorted_references = convert_verse_ids_to_references(verse_ids)

    formatted_reference = ""

    previous_reference = None

    for reference in sorted_references:
        book, start_chapter, start_verse, end_chapter, end_verse = reference

        # First reference
        if previous_reference is None:
            formatted_reference += format_single_reference(
                book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            previous_reference = reference
            continue

        previous_book = previous_reference[0]

        # Reference with a new book
        if previous_book != book:
            formatted_reference += ";"
            formatted_reference += format_single_reference(
                book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            previous_reference = reference
            continue

        previous_end_chapter = previous_reference[3]

        # Reference with a new chapter
        if previous_end_chapter != start_chapter or end_chapter > start_chapter:
            formatted_reference += ","
            formatted_reference += format_single_reference(
                None, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            continue

        # Reference with same book and chapter as previous reference
        formatted_reference += ","
        formatted_reference += format_single_reference(
            None, None, start_verse, None, end_verse, **kwargs
        )
        previous_reference = reference

    return formatted_reference


def format_single_reference(
    book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
):
    formatted_reference = ""

    if book:
        parser = kwargs.get("parser", DEFAULT_PARSER)
        full_title = kwargs.get("full_title", False)

        title = (
            parser.get_book_title(book)
            if full_title
            else parser.get_short_book_title(book)
        )

        formatted_reference += f"{title} "

    if start_chapter:
        formatted_reference += f"{start_chapter}:{start_verse}"
    else:
        formatted_reference += f"{start_verse}"

    if end_chapter and end_chapter > start_chapter:
        formatted_reference += f"-{end_chapter}:{end_verse}"
    elif end_verse > start_verse:
        formatted_reference += f"-{end_verse}"

    return formatted_reference


def format_scripture_text(verse_ids, **kwargs):
    parser = kwargs.get("parser", DEFAULT_PARSER)
    full_title = kwargs.get("full_title", False)
    title_function = (
        parser.get_book_title if full_title else parser.get_short_book_title
    )
    format_type = kwargs.get("format_type", "html")
    text = ""

    paragraphs = parser.get_scripture_passage_text(verse_ids, **kwargs)

    for book, chapters in paragraphs.items():
        title = title_function(book)

        if format_type == "html":
            text += f"<h1>{title}</h1>\n"
        else:
            if len(text) > 0:
                text += "\n\n"

            text += f"{title}\n\n"

        for chapter, paragraphs in chapters.items():
            if format_type == "html":
                text += f"<h2>Chapter {chapter}</h2>\n"
            else:
                text += f"Chapter {chapter}\n\n"

            for paragraph in paragraphs:
                if format_type == "html":
                    text += f"<p>{paragraph}</p>\n"
                else:
                    text += f"   {paragraph}\n"

    return text
